﻿using System;
using System.IO;

namespace JDB
{
    // Without any sea-green stuff! :D
    // Version 1.0
    // Created by joppiesaus, http://joppiesaus.function1.nl/, view Documentation.txt for more info!
    public class DataBase
    {
        private string[][] lines;
        public string[][] Rows { get { return lines; } }
        
        private string[] l;
        public string[] Lines { get { return l; } }

        private string path;
        public string Path { get { return path; } }

        private string[] columns;
        public string[] Columns { get { return columns; } }

        private char splitChar;
        public char SplitCharacter { get { return splitChar; } }

        private char hookChar;
        public char HookCharacter { get { return hookChar; } }

        public int RowLength { get { return l.Length; } }
        public int ColumnLength { get { return columns.Length; } }

        public DataBase(string dbpath, string[] data, char splitCharacter = ' ', char hookCharacter = '"')
        {
            this.path = Directory.GetCurrentDirectory() + "\\" + dbpath;
            this.columns = data;
            this.splitChar = splitCharacter;
            this.hookChar = hookCharacter;
            readFile();
        }

        // SELECT * FROM people WHERE clue > 0
        // 0 rows returned
        public string[][] getRows(int index = -1, string arg = "", bool ignoreCase = false, bool getContaining = false)
        {
            return getRows(lines, columns.Length, index, arg, ignoreCase, getContaining);
        }

        public string[][] getRows(string column = "", string arg = "", bool ignoreCase = false, bool getContaining = false)
        {
            return getRows(getIndex(column, ignoreCase), arg, ignoreCase, getContaining);
        }

        public string get(int dimension1, int dimension2)
        {
            if (dimension1 >= 0 && dimension1 < lines.Length && dimension2 >= 0 && dimension2 < columns.Length) return lines[dimension1][dimension2];
            return null;
        }

        public string[] getRow(int row)
        {
            if (row >= 0 && row < lines.Length) return lines[row];
            return null;
        }

        public bool exists(int index = -1, string arg = "", bool ignoreCase = false, bool containing = false)
        {
            return (getRows(index, arg, ignoreCase, containing) != null);
        }

        public bool exists(string column = "", string arg = "", bool ignoreCase = false, bool containing = false)
        {
            return exists(getIndex(column, ignoreCase), arg, ignoreCase, containing);
        }

        private static bool isInArray(int[] data, int line)
        {
            for (int i = 0; i < data.Length; i++)
            {
                //if (data[i] == null) break;
                if (data[i] == line) return true;
            }
            return false;
        }

        public string getWriteString(string[] data)
        {
            if (data.Length != columns.Length) return null;

            int a = 0;
            string[] parts = new string[data.Length];

            for (int i = 0; i < data.Length; i++)
            {
                parts[a] = hookChar + data[i].Replace("" + hookChar, "#hookchar#").Replace(Environment.NewLine, "#nlchar#") + hookChar + splitChar;
                a++;
            }

            parts[a - 1] = parts[a - 1].Split(splitChar)[0]; // Our last problem, is now gone forever! Muahahahahah!

            string finalString = "";
            for (int i = 0; i < parts.Length; i++)
            {
                finalString += parts[i];
            }
            return finalString;
        }

        // Just for the lazy people like me
        public void add(object[] data)
        {
            string[] s = new string[data.Length];
            for (int i = 0; i < data.Length; i++)
            {
                s[i] = data[i].ToString();
            }
            add(s);
        }

        // Well this was once bigger.
        public void add(string[] data)
        {
            writeToDatabaseFile(getWriteString(data), true);
        }

        private void rem(int where)
        {
            for (int i = where + 1; i < lines.Length; i++)
            {
                lines[i - 1] = lines[i]; // lie
            }
            if (lines.Length > 0) Array.Resize(ref lines, lines.Length - 1);
        }

        public void edit(int where, string[] data)
        {
            if (where < 0 || where >= lines.Length || data.Length != columns.Length) return;
            
            lines[where] = data;

            checkFile();

            string readed = "";
            using (StreamReader read = new StreamReader(path)) readed = read.ReadToEnd();

            string[] temp = readed.Split(new string[] { Environment.NewLine }, StringSplitOptions.None);
            temp[where] = getWriteString(data);

            string finalString = "";
            for(int i = 0; i < temp.Length - 1; i++)
            {
                finalString += temp[i] + Environment.NewLine;
            }
            if (temp.Length > 0) finalString += temp[temp.Length - 1];

            using (StreamWriter write = new StreamWriter(path)) write.Write(finalString);
        }

        // Refresh-like method
        public void writeToDatabaseFile(string extra = "", bool readAfterWrite = false)
        {
            checkFile();

            string readed = "";
            using (StreamReader read = new StreamReader(path)) readed = read.ReadToEnd();

            if (readed != "") readed += Environment.NewLine;
            using (StreamWriter write = new StreamWriter(path)) write.Write(readed + extra);

            if (readAfterWrite) readFile(); // Re-init and done!
        }

        public void rewriteDatabase()
        {
            string toWrite = "";
            for (int i = 0; i < lines.Length - 1; i++)
            {
                toWrite += getWriteString(lines[i]) + Environment.NewLine;
            }
            toWrite += getWriteString(lines[lines.Length - 1]);

            checkFile();
            using (StreamWriter write = new StreamWriter(path)) write.Write(toWrite);
        }

        public void edit(int[] where, string[][] data)
        {
            for (int i = 0; i < where.Length; i++) edit(where[i], data[i]);
        }

        public void remove(int where)
        {
            rem(where);
            rewriteDatabase();
        }

        public void remove(int[] data)
        {
            for (int i = 0; i < data.Length; i++) rem(data[i]);
            rewriteDatabase();
        }

        public int getIndex(string column, bool ignoreCase = false)
        {
            if (ignoreCase)
            {
                if (column != null) column = column.ToLower();
                for (int i = 0; i < columns.Length; i++)
                {
                    if (columns[i].ToLower() == column) return i;
                }
            }
            else
            {
                for (int i = 0; i < columns.Length; i++)
                {
                    if (columns[i] == column) return i;
                }
            }
            return -1;
        }

        // returns ONLY the first one, and reacts ONLY if the value is EXACT the same.
        public int getRowIndex(string[] data)
        {
            for (int i = 0; i < lines.Length; i++)
            {
                if (lines[i] == data) return i;
            }
            return -1;
        }

        public string getColumn(int index)
        {
            return columns[index];
        }

        public void checkFile()
        {
            if (!File.Exists(path)) using (File.Create(path)) { }
        }

        public void readFile()
        {
            checkFile();

            using (StreamReader r = new StreamReader(path))
            {
                l = r.ReadToEnd().Split(new string[] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries);
            }

            lines = new string[l.Length][];

            for (int i = 0; i < lines.Length; i++)
            {
                lines[i] = getReadingString(l[i].Split(hookChar));
                l[i] = l[i];
            }
        }

        private string[] getReadingString(string[] data)
        {
            string[] toRet = new string[data.Length];
            try
            {
                int a = 0;
                for (int i = 1; i < data.Length; i += 2)
                {
                    toRet[a] = data[i].Replace("#hookchar#", hookChar.ToString()).Replace("#nlchar#", Environment.NewLine);
                    a++;
                }
                Array.Resize(ref toRet, a);
                return toRet;
            }
            catch // Will cause holes, so don't mess it up.
            {
                return null;
            }
        }

        // ListBoxes can be a pain in the ass.
        public string[] getListBoxStrings(int dimension = 0)
        {
            return getListBoxStrings(dimension, lines);
        }

        public string[] getListBoxStrings(int dimension, string[][] data)
        {
            if (dimension > columns.Length || dimension < 0) dimension = 0;
            string[] toRet = new string[data.Length];
            int a = 0;

            for (int i = 0; i < data.Length; i++)
            {
                toRet[a] = data[i][dimension];
                a++;
            }

            Array.Resize(ref toRet, a);
            return toRet;
        }

        // You could make this static if you want to
        // I did this because with this you can get search results within search results within... SEARCHCEPTION!
        public string[][] getRows(string[][] data, int columnLength, int index = -1, string arg = "", bool ignoreCase = false, bool getContaining = false)
        {
            if (index >= columnLength) return null; // I love null!
            if (!getContaining && arg == "") return data;

            string[][] toReturn = new string[data.Length][];
            string[][] lns = new string[data.Length][];
            int a = 0;

            // copy it manually. http://www.yoda.arachsys.com/csharp/references.html
            // Also, I HATE doing this again. But it costs less performance, so I do this.
            if (ignoreCase) for (int i = 0; i < lns.Length; i++)
            {
                arg = arg.ToLower();
                lns[i] = new string[data[i].Length];
                for (int j = 0; j < lns[i].Length; j++)
                    lns[i][j] = data[i][j].ToLower();
            }
            else for (int i = 0; i < lns.Length; i++)
            {
                lns[i] = new string[data[i].Length];
                for (int j = 0; j < lns[i].Length; j++)
                    lns[i][j] = data[i][j];
            }



            if (index >= 0)
            {
                if (getContaining)
                {
                    for (int i = 0; i < data.Length; i++)
                        if (lns[i][index].Contains(arg))
                        {
                            toReturn[a] = data[i];
                            a++;
                        }
                }
                else for (int i = 0; i < data.Length; i++)
                        if (lns[i][index] == arg)
                        {
                            toReturn[a] = data[i];
                            a++;
                        }
            }
            else // UGH! I really hate to write the same code multiple times!
            {
                int[] thingsInArray = new int[data.Length];
                int b = 0;

                for (int i = 0; i < thingsInArray.Length; i++)
                    thingsInArray[i] = -1;

                if (getContaining) for (int i = 0; i < columnLength; i++)
                        for (int j = 0; j < data.Length; j++)
                        {
                            if (lns[j][i].Contains(arg))
                            {
                                if (isInArray(thingsInArray, j)) continue;
                                thingsInArray[b] = j;
                                b++;
                                toReturn[a] = data[j];
                                a++;
                            }
                        }
                else for (int i = 0; i < columnLength; i++)
                        for (int j = 0; j < data.Length; j++)
                            if (lns[j][i] == arg)
                            {
                                if (isInArray(thingsInArray, j)) continue;
                                thingsInArray[b] = j;
                                b++;
                                toReturn[a] = data[j];
                                a++;
                            }
            }

            if (a > 0) Array.Resize(ref toReturn, a);
            else return null;

            return toReturn;
        }
    }

    public static class StringHelper // He doesn't get paid sadly enough.
    {
        // Handy extension method if you want to display a row quick without everything
        public static string toPhraseString(this string[] s)
        {
            string toReturn = "";
            for (int i = 0; i < s.Length - 1; i++)
            {
                toReturn += s[i] + " ";
            }
            if (s.Length > 0) toReturn += s[s.Length - 1];
            return toReturn;
        }
    }
}
